import { As } from "@kobalte/core";
import { ComponentProps, Match, Switch } from "solid-js";

import * as Icons from "@/icons";
import { Tooltip, TooltipContent, TooltipTrigger } from "./ui/tooltip";

import { setStore, store } from "@/contexts/player";

import player from "@/lib/player";
import { formatDuration } from "@/lib/duration";
import { parseDuration } from "@/lib/time";
import { A } from "@solidjs/router";
import { cn } from "@/lib/utils";

function VolumeIcon() {
	return (
		<Switch>
			<Match when={store.muted}>
				<Icons.VolumeOffStroke class="size-6" />
			</Match>
			<Match when={store.volume <= 0}>
				<Icons.VolumeMute02Stroke class="size-6" />
			</Match>
			<Match when={store.volume <= 0.33}>
				<Icons.VolumeMute01Stroke class="size-6" />
			</Match>
			<Match when={store.volume <= 0.66}>
				<Icons.VolumeLowStroke class="size-6" />
			</Match>
			<Match when={store.volume <= 1}>
				<Icons.VolumeHighStroke class="size-6" />
			</Match>
		</Switch>
	);
}

function PlayPauseIcon() {
	return (
		<Switch>
			<Match when={store.playing}>
				<Icons.PauseStroke class="size-8 text-primary" />
			</Match>
			<Match when={!store.playing}>
				<Icons.PlayStroke class="size-8 text-primary" />
			</Match>
		</Switch>
	);
}

function RepeatIcon() {
	return (
		<Switch fallback={<Icons.RepeatStroke class="size-4 text-secondary" />}>
			<Match when={store.repeat === "all"}>
				<Icons.RepeatStroke class="size-4 text-primary" />
			</Match>
			<Match when={store.repeat === "one"}>
				<Icons.RepeatOne01Stroke class="size-4 text-primary" />
			</Match>
		</Switch>
	);
}

export default function Player() {
	return (
		<div class="space grid h-24 grid-cols-3 gap-6">
			<div class="flex items-center gap-4">
				<div class="aspect-square h-full rounded-md border bg-secondary" />
				<div class="flex flex-1 flex-col gap-1">
					<h4>{store.track.title}</h4>
					<span class="text-sm text-secondary">
						{store.track.artists.map((artist, i) => (
							<>
								<A class="hover:underline" href={`/artists/${artist.id}`}>
									{artist.name}
								</A>
								{i < store.track.artists.length - 1 && ", "}
							</>
						))}
					</span>
				</div>
			</div>

			<div class="flex flex-col items-center justify-center gap-2">
				<div class="flex items-center gap-4">
					<Button
						tooltip="Shuffle"
						onClick={() =>
							setStore((store) => ({ ...store, shuffle: !store.shuffle }))
						}
						class={cn({
							"bg-tertiary text-primary": store.shuffle,
						})}
					>
						<Icons.ShuffleStroke class="size-4" />
					</Button>

					<Button tooltip="Previous">
						<Icons.PreviousStroke class="size-6" />
					</Button>

					<Button
						tooltip={store.playing ? "Pause" : "Play"}
						onClick={() => player.playPause()}
					>
						<PlayPauseIcon />
					</Button>

					<Button tooltip="Next">
						<Icons.NextStroke class="size-6" />
					</Button>

					<Button
						tooltip="Repeat"
						onClick={() =>
							setStore((store) => ({
								...store,
								repeat:
									store.repeat === "none"
										? "all"
										: store.repeat === "all"
											? "one"
											: "none",
							}))
						}
						class={cn({
							"bg-tertiary text-primary":
								store.repeat === "all" || store.repeat === "one",
						})}
					>
						<RepeatIcon />
					</Button>
				</div>

				<div class="flex w-full items-center gap-4 text-sm">
					<span class="w-16">
						{formatDuration(parseDuration(store.progress))}
					</span>
					<div
						class="h-1.5 w-full cursor-pointer overflow-hidden rounded-md bg-tertiary"
						onClick={(event) => {
							const rect = event.currentTarget.getBoundingClientRect();
							const x = event.clientX - rect.left;
							const width = rect.width;
							const progress = x / width;

							player.seek(progress * store.duration);
						}}
					>
						<div
							class="h-full w-0 rounded-md bg-accent"
							style={{ width: `${(store.progress / store.duration) * 100}%` }}
						/>
					</div>
					<span class="w-16 text-right">
						{formatDuration(parseDuration(store.duration))}
					</span>
				</div>
			</div>

			<div class="ml-auto flex items-center gap-4">
				<Button tooltip="Sound output">
					<Icons.RadioStroke class="size-6" />
				</Button>

				<Button
					tooltip="Queue"
					onClick={() => {
						setStore((store) => ({ ...store, queueOpen: !store.queueOpen }));
					}}
				>
					<Icons.Queue02Stroke class="size-6" />
				</Button>

				<div class="flex items-center gap-2">
					<Button
						tooltip="Volume"
						onClick={() =>
							setStore((store) => ({ ...store, muted: !store.muted }))
						}
					>
						<VolumeIcon />
					</Button>

					<input
						type="range"
						class="block h-full max-w-24 accent-accent"
						value={store.muted ? 0 : store.volume}
						onInput={(event) => {
							setStore((store) => ({
								...store,
								volume: Number(event.currentTarget.value),
								muted: false,
							}));
						}}
						min={0}
						step={0.01}
						max={1}
					/>
				</div>
			</div>
		</div>
	);
}

interface ButtonProps extends ComponentProps<"button"> {
	tooltip?: string;
	children: any;

	onClick?: () => void;
}
function Button(props: ButtonProps) {
	return (
		<Tooltip placement="top">
			<TooltipTrigger asChild>
				<As
					component="button"
					class={cn(
						"flex size-8 cursor-pointer items-center justify-center rounded-md text-secondary transition-colors hover:scale-105 hover:text-primary",
						props.class,
					)}
					onClick={props.onClick}
				>
					{props.children}
				</As>
			</TooltipTrigger>
			<TooltipContent>{props.tooltip}</TooltipContent>
		</Tooltip>
	);
}
