import ChunkQueue from "./chunk-queue";
import type { Events, EventListeners } from "./types/player";

import { DEFAULT_PLAYBACK_INFO, PlaybackInfo } from "./types/playback-info";
import { DEFAULT_TRACK, type Track } from "@/types/track";
import chunkCache from "./chunk-cache";

const CHUNK_LENGTH = 5,
	PRELOAD_CHUNKS = 3;

export class Player {
	static audio = new Audio();

	private _listeners = new Map<Events, EventListeners[Events][]>();
	private _initialized = false;

	private track = DEFAULT_TRACK;

	private queue = new ChunkQueue(this);

	private playbackInfo: PlaybackInfo = DEFAULT_PLAYBACK_INFO;
	// private state

	private sourceBuffer: SourceBuffer | null = null;

	currentChunk = 0;

	constructor() {
		navigator.mediaSession.setActionHandler("pause", () => this.playPause());
		navigator.mediaSession.setActionHandler("play", () => this.playPause());

		this.createMediaSource = this.createMediaSource.bind(this);
		this.sourceOpen = this.sourceOpen.bind(this);

		// this.startStateUpdate = this.startStateUpdate.bind(this);
		// this.updateState = this.updateState.bind(this);

		this.reset = this.reset.bind(this);
		this.play = this.play.bind(this);
		this.playPause = this.playPause.bind(this);
		this.runQueue = this.runQueue.bind(this);

		this.changeTrack = this.changeTrack.bind(this);
		this.processNextChunk = this.processNextChunk.bind(this);

		this.registerEvents = this.registerEvents.bind(this);

		this.registerEvents();
	}

	private registerEvents() {
		this.on("queue", this.runQueue);
	}

	getAudio() {
		return Player.audio;
	}

	getTrack(): Track {
		return this.track;
	}

	/**
	 * Change track.
	 */
	async changeTrack(track: Track, progress = 0) {
		// If track is the same, just play/pause
		if (this.track.id === track.id && !progress) {
			this.playPause();
			return;
		}

		this.reset();

		Player.audio.pause();

		if (this.track.id !== track.id) {
			this.queue.clear();
		}

		this.track = track;
		await this.fetchPlaybackInfo();

		this.emit("track", track);

		// Update mediaSession
		if ("mediaSession" in navigator) {
			navigator.mediaSession.metadata = new MediaMetadata({
				title: track.title,
				artist: track.artists.map((artist) => artist.name).join(", "),
				album: "Unknown",
				// artwork: [
				// 	{
				// 		src: track.image,
				// 		sizes: "1280x1280",
				// 		type: "image/jpg",
				// 	},
				// ],
			});
		}

		this.queue = new ChunkQueue(this);

		this._progress = progress;
		this.currentChunk = getChunkFromTime(progress);

		this.play();
	}
	private _progress = 0;

	private _checkedSecond = 0;
	private processNextChunk() {
		const second = Math.floor(Player.audio.currentTime);

		if (second % CHUNK_LENGTH !== 0 || this._checkedSecond === second) {
			return;
		}

		this._checkedSecond = second;

		const nextChunk = this.currentChunk + 1;
		this.queue.setIndex(nextChunk);

		if (nextChunk > this.playbackInfo.max_chunks) {
			this.sourceBuffer?.removeEventListener("updateend", this.runQueue);
			return;
		}

		this.queue.load(nextChunk);

		this.currentChunk = nextChunk;
	}

	private async fetchPlaybackInfo() {
		const { id } = this.track;

		const data = await fetch(`http://localhost:9001/tracks/${id}`).then((res) =>
			res.json(),
		);

		this.playbackInfo = data;
	}

	private createMediaSource() {
		const mediaSource = new MediaSource();

		const audio = this.getAudio();
		audio.src = URL.createObjectURL(mediaSource);

		audio.currentTime = this._progress;

		mediaSource.addEventListener("sourceopen", () => {
			this.sourceOpen(mediaSource);
		});
	}

	private async sourceOpen(mediaSource: MediaSource) {
		const { queue } = this;
		queue.clear();
		chunkCache.clear();

		this.sourceBuffer = mediaSource.addSourceBuffer(this.getSourceBufferType());

		--this.currentChunk;

		this.queue.setIndex(0);
		await this.queue.load(0);

		this.queue.setIndex(++this.currentChunk);

		// Prleoad x chunks.
		for (let i = 0; i < PRELOAD_CHUNKS; i++) {
			if (this.currentChunk > this.playbackInfo.max_chunks) {
				break;
			}

			// Fetch chunk.
			this.queue.load(this.currentChunk++);
		}

		--this.currentChunk;

		this.sourceBuffer.addEventListener("updateend", this.runQueue);

		Player.audio.addEventListener("timeupdate", this.processNextChunk);
	}

	private getSourceBufferType(): string {
		const types = ["audio/mp4", 'audio/mp4; codecs="flac"'];

		for (const type of types) {
			if (MediaSource.isTypeSupported(type)) {
				return type;
			}
		}

		return "";
	}

	private runQueue() {
		const { queue } = this;

		if (queue.size === 0 || this.sourceBuffer?.updating) {
			return;
		}

		const buffer = queue.next();

		if (!buffer) {
			return;
		}

		try {
			this.sourceBuffer?.appendBuffer(buffer);
		} catch (error) {
			console.error(error);
		}
	}

	private play() {
		if (!this._initialized) {
			this.createMediaSource();

			this._initialized = true;
		}

		Player.audio.play();
	}

	playPause() {
		if (!this.track.id) return;

		const audio = this.getAudio();

		if (audio.paused) {
			this.play();
		} else {
			audio.pause();
		}
	}

	private reset() {
		this._initialized = false;

		if (this.sourceBuffer) {
			this.sourceBuffer.removeEventListener("updateend", this.runQueue);
		}

		Player.audio.removeEventListener("timeupdate", this.processNextChunk);
	}

	public getPlaybackInfo() {
		return this.playbackInfo;
	}

	/// Controls
	get volume() {
		return Player.audio.volume;
	}

	set volume(value: number) {
		Player.audio.volume = value;
	}

	get muted() {
		return Player.audio.muted;
	}

	set muted(value: boolean) {
		Player.audio.muted = value;
	}

	get currentTime() {
		return Player.audio.currentTime;
	}

	isPlaying() {
		return !Player.audio.paused;
	}

	seek(progress: number) {
		this.changeTrack(this.track, progress);
	}

	/// Events
	public emit<E extends Events>(event: E, ...args: unknown[]) {
		const listeners: EventListeners[Events][] =
			this._listeners.get(event) ?? [];

		// eslint-disable-next-line @typescript-eslint/no-unsafe-call, @typescript-eslint/ban-ts-comment
		// @ts-expect-error
		listeners.forEach((listener) => listener(...args));
	}

	public on<E extends Events>(event: E, listener: EventListeners[E]) {
		const eventListeners = this._listeners.get(event) ?? [];

		eventListeners.push(listener);

		this._listeners.set(event, eventListeners);
	}

	public off<E extends Events>(event: E, listener: EventListeners[E]) {
		const eventListeners = this._listeners.get(event) ?? [];

		this._listeners.set(
			event,
			eventListeners.filter((eventListener) => eventListener !== listener),
		);
	}
}

const player = new Player();

export default player;

function getChunkFromTime(time: number) {
	return Math.floor(time / CHUNK_LENGTH);
}
