import { A, useParams } from "@solidjs/router";
import { For, Show, createResource, createSignal, useContext } from "solid-js";

import { fetchAPI } from "@/lib/api";
import { Playlist } from "@/types/playlist";
import NotFound from "../not-found";
import { UserContext } from "@/contexts/user";
import { cn } from "@/lib/utils";

export default function User() {
	const { user: loggedInUser } = useContext(UserContext);
	const params = useParams<{ userId: string }>();

	const [user] = createResource(() => params.userId, getUser);
	const [playlists] = createResource(() => params.userId, getUserPlaylists);

	const isSelf = loggedInUser.id === user()?.id;

	const [followed, setFollowed] = createSignal(false);

	return (
		<Show
			when={!user.loading}
			fallback={
				<Show
					when={user.loading}
					fallback={<NotFound />}
					children={<p>Loading...</p>}
				/>
			}
		>
			<section>
				<header class="mb-8 flex items-center gap-4 border-b pb-8">
					<div class="aspect-square size-40 rounded-full bg-tertiary xl:size-48" />
					<div class="flex flex-col items-start gap-4">
						<h1 class="text-3xl font-bold">{user()?.name}</h1>
						<p class="text-sm uppercase tracking-widest text-secondary">
							x Followers - x Following
						</p>
					</div>

					<button
						class={cn(
							"ml-auto rounded-md border bg-primary px-4 py-2 text-secondary transition-colors hover:bg-secondary hover:text-primary",
							{
								"bg-white text-secondary": followed(),
							},
						)}
						onClick={() => setFollowed(!followed())}
					>
						{followed() ? "Unfollow" : "Follow"}
					</button>
				</header>

				<div class="flex flex-col gap-8 md:gap-16">
					<div>
						<h2 class="mb-4 text-2xl font-bold">Public Playlists</h2>
						<div class="grid grid-cols-2 gap-8 md:grid-cols-3 lg:grid-cols-4 xl:grid-cols-5">
							<For each={playlists()}>
								{(playlist) => (
									<A
										href={`/playlists/${playlist.id}`}
										class="flex flex-col overflow-hidden rounded-lg border bg-primary transition-transform hover:scale-105"
									>
										<div class="aspect-square w-full bg-tertiary" />

										<div class="p-2">
											<h4 class="text-lg font-bold">{playlist.name}</h4>
											<p class="line-clamp-2 text-sm leading-tight text-secondary">
												Lorem ipsum dolor sit amet consectetur consectetur
												consectetur
											</p>
										</div>
									</A>
								)}
							</For>
						</div>
					</div>

					<div>
						<h2 class="mb-4 text-2xl font-bold">Followers</h2>
						<div class="grid grid-cols-2 gap-8 md:grid-cols-3 lg:grid-cols-4 xl:grid-cols-5">
							<For each={playlists()}>
								{(playlist) => (
									<A
										href={`/playlists/${playlist.id}`}
										class="flex flex-col overflow-hidden rounded-lg border bg-primary transition-transform hover:scale-105"
									>
										<div class="aspect-square w-full bg-tertiary" />

										<div class="p-2">
											<h4 class="text-lg font-bold">{playlist.name}</h4>
											<p class="line-clamp-2 text-sm leading-tight text-secondary">
												Lorem ipsum dolor sit amet consectetur consectetur
												consectetur
											</p>
										</div>
									</A>
								)}
							</For>
						</div>
					</div>

					<div>
						<h2 class="mb-4 text-2xl font-bold">Following</h2>
						<div class="grid grid-cols-2 gap-8 md:grid-cols-3 lg:grid-cols-4 xl:grid-cols-5">
							<For each={playlists()}>
								{(playlist) => (
									<A
										href={`/playlists/${playlist.id}`}
										class="flex flex-col overflow-hidden rounded-lg border bg-primary transition-transform hover:scale-105"
									>
										<div class="aspect-square w-full bg-tertiary" />

										<div class="p-2">
											<h4 class="text-lg font-bold">{playlist.name}</h4>
											<p class="line-clamp-2 text-sm leading-tight text-secondary">
												Lorem ipsum dolor sit amet consectetur consectetur
												consectetur
											</p>
										</div>
									</A>
								)}
							</For>
						</div>
					</div>
				</div>
			</section>
		</Show>
	);
}

const users = new Map<string, UserType>();

async function getUser(userId: string): Promise<UserType> {
	if (users.has(userId)) {
		return users.get(userId)!;
	}

	const user = await fetchAPI<UserType>(`/users/${userId}`);

	users.set(userId, user);

	return user;
}

async function getUserPlaylists(userId: string): Promise<Playlist[]> {
	return fetchAPI(`/users/${userId}/playlists`);
}

interface UserType {
	id: string;
	name: string;
}
